@@ -1,6 +1,8 @@
 /*
 
-Copyright (c) 2003-2016, Arvid Norberg
+Copyright (c) 2003-2004, 2006-2010, 2012, 2014-2020, Arvid Norberg
+Copyright (c) 2016-2017, Alden Torres
+Copyright (c) 2018, Alexandre Janniaux
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without
@@ -34,6 +36,7 @@ POSSIBILITY OF SUCH DAMAGE.
 #define TORRENT_SOCKET_HPP_INCLUDED
 
 #include "libtorrent/config.hpp"
+#include "libtorrent/aux_/noexcept_movable.hpp"
 
 #include "libtorrent/aux_/disable_warnings_push.hpp"
 
@@ -50,8 +53,6 @@ POSSIBILITY OF SUCH DAMAGE.
 #include <winsock2.h>
 #endif
 
-#include <boost/version.hpp>
-
 #include <boost/asio/ip/tcp.hpp>
 #include <boost/asio/ip/udp.hpp>
 #include <boost/asio/write.hpp>
@@ -65,25 +66,56 @@ POSSIBILITY OF SUCH DAMAGE.
 #include "simulator/simulator.hpp"
 #endif
 
+#if TORRENT_USE_NETLINK
+#include <linux/netlink.h>
+#ifndef SOL_NETLINK
+#define SOL_NETLINK 270
+#endif
+
+// NETLINK_NO_ENOBUFS exists at least since android 2.3, but is not exposed
+#if defined TORRENT_ANDROID && !defined NETLINK_NO_ENOBUFS
+#define NETLINK_NO_ENOBUFS 5
+#endif
+#endif
+
 #include "libtorrent/aux_/disable_warnings_pop.hpp"
 
-namespace libtorrent
-{
+namespace libtorrent {
 
 #if defined TORRENT_BUILD_SIMULATOR
-	using sim::asio::ip::udp;
-	using sim::asio::ip::tcp;
+struct tcp : sim::asio::ip::tcp {
+	tcp(sim::asio::ip::tcp const& p) : sim::asio::ip::tcp(p) {} // NOLINT
+	using socket = aux::noexcept_move_only<sim::asio::ip::tcp::socket>;
+};
+struct udp : sim::asio::ip::udp {
+	udp(sim::asio::ip::udp const& p) : sim::asio::ip::udp(p) {} // NOLINT
+	using socket = aux::noexcept_move_only<sim::asio::ip::udp::socket>;
+};
 	using sim::asio::async_write;
 	using sim::asio::async_read;
-	using sim::asio::null_buffers;
+	using true_tcp_socket = sim::asio::ip::tcp::socket;
 #else
-	using boost::asio::ip::tcp;
-	using boost::asio::ip::udp;
+struct tcp : boost::asio::ip::tcp {
+	tcp(boost::asio::ip::tcp const& p) : boost::asio::ip::tcp(p) {} // NOLINT
+	using socket = aux::noexcept_move_only<boost::asio::ip::tcp::socket>;
+};
+struct udp : boost::asio::ip::udp {
+	udp(boost::asio::ip::udp const& p) : boost::asio::ip::udp(p) {} // NOLINT
+	using socket = aux::noexcept_move_only<boost::asio::ip::udp::socket>;
+};
 	using boost::asio::async_write;
 	using boost::asio::async_read;
-	using boost::asio::null_buffers;
+	using true_tcp_socket = boost::asio::ip::tcp::socket;
 #endif
 
+	// internal
+	inline udp::endpoint make_udp(tcp::endpoint const ep)
+	{ return {ep.address(), ep.port()}; }
+
+	// internal
+	inline tcp::endpoint make_tcp(udp::endpoint const ep)
+	{ return {ep.address(), ep.port()}; }
+
 #ifdef TORRENT_WINDOWS
 
 #ifndef PROTECTION_LEVEL_UNRESTRICTED
@@ -96,7 +128,7 @@ namespace libtorrent
 
 	struct v6_protection_level
 	{
-		v6_protection_level(int level): m_value(level) {}
+		explicit v6_protection_level(int level): m_value(level) {}
 		template<class Protocol>
 		int level(Protocol const&) const { return IPPROTO_IPV6; }
 		template<class Protocol>
@@ -110,7 +142,7 @@ namespace libtorrent
 
 	struct exclusive_address_use
 	{
-		exclusive_address_use(int enable): m_value(enable) {}
+		explicit exclusive_address_use(int enable): m_value(enable) {}
 		template<class Protocol>
 		int level(Protocol const&) const { return SOL_SOCKET; }
 		template<class Protocol>
@@ -126,7 +158,7 @@ namespace libtorrent
 #ifdef IPV6_TCLASS
 	struct traffic_class
 	{
-		traffic_class(char val): m_value(val) {}
+		explicit traffic_class(int val): m_value(val) {}
 		template<class Protocol>
 		int level(Protocol const&) const { return IPPROTO_IPV6; }
 		template<class Protocol>
@@ -142,11 +174,11 @@ namespace libtorrent
 	struct type_of_service
 	{
 #ifdef _WIN32
-		typedef DWORD tos_t;
+		using tos_t = DWORD;
 #else
-		typedef int tos_t;
+		using tos_t = int;
 #endif
-		type_of_service(char val): m_value(val) {}
+		explicit type_of_service(tos_t const val) : m_value(tos_t(val)) {}
 		template<class Protocol>
 		int level(Protocol const&) const { return IPPROTO_IP; }
 		template<class Protocol>
@@ -158,6 +190,22 @@ namespace libtorrent
 		tos_t m_value;
 	};
 
+#ifdef IP_DSCP_TRAFFIC_TYPE
+	struct dscp_traffic_type
+	{
+		explicit dscp_traffic_type(DWORD val) : m_value(val) {}
+		template<class Protocol>
+		int level(Protocol const&) const { return IP_DSCP_TRAFFIC_TYPE; }
+		template<class Protocol>
+		int name(Protocol const&) const { return DSCP_TRAFFIC_TYPE; }
+		template<class Protocol>
+		DWORD const* data(Protocol const&) const { return &m_value; }
+		template<class Protocol>
+		size_t size(Protocol const&) const { return sizeof(m_value); }
+		DWORD m_value;
+	};
+#endif
+
 #if defined IP_DONTFRAG || defined IP_MTU_DISCOVER || defined IP_DONTFRAGMENT
 #define TORRENT_HAS_DONT_FRAGMENT
 #endif
@@ -172,7 +220,7 @@ namespace libtorrent
 
 	struct dont_fragment
 	{
-		dont_fragment(bool val) : m_value(val) {}
+		explicit dont_fragment(bool val) : m_value(val) {}
 		template<class Protocol>
 		int level(Protocol const&) const { return IPPROTO_IP; }
 		template<class Protocol>
@@ -197,8 +245,8 @@ namespace libtorrent
 	// it.
 	struct dont_fragment
 	{
-		dont_fragment(bool val)
-			: m_value(val ? IP_PMTUDISC_DO : IP_PMTUDISC_DONT) {}
+		explicit dont_fragment(bool val)
+			: m_value(val ? IP_PMTUDISC_PROBE : IP_PMTUDISC_DONT) {}
 		template<class Protocol>
 		int level(Protocol const&) const { return IPPROTO_IP; }
 		template<class Protocol>
@@ -213,7 +261,38 @@ namespace libtorrent
 #endif // IP_DONTFRAG vs. IP_MTU_DISCOVER
 
 #endif // TORRENT_HAS_DONT_FRAGMENT
+
+#if TORRENT_USE_NETLINK
+	struct no_enobufs
+	{
+		explicit no_enobufs(bool val) : m_value(val) {}
+		template<class Protocol>
+		int level(Protocol const&) const { return SOL_NETLINK; }
+		template<class Protocol>
+		int name(Protocol const&) const { return NETLINK_NO_ENOBUFS; }
+		template<class Protocol>
+		int const* data(Protocol const&) const { return &m_value; }
+		template<class Protocol>
+		std::size_t size(Protocol const&) const { return sizeof(m_value); }
+		int m_value;
+	};
+#endif // TORRENT_USE_NETLINK
+
+#ifdef TCP_NOTSENT_LOWAT
+	struct tcp_notsent_lowat
+	{
+		explicit tcp_notsent_lowat(int val) : m_value(val) {}
+		template<class Protocol>
+		int level(Protocol const&) const { return IPPROTO_TCP; }
+		template<class Protocol>
+		int name(Protocol const&) const { return TCP_NOTSENT_LOWAT; }
+		template<class Protocol>
+		int const* data(Protocol const&) const { return &m_value; }
+		template<class Protocol>
+		std::size_t size(Protocol const&) const { return sizeof(m_value); }
+		int m_value;
+	};
+#endif
 }
 
 #endif // TORRENT_SOCKET_HPP_INCLUDED
-
